#!/usr/bin/env ruby
require "socket"
require "singleton"


module ConfigurationHelper
  def def_conf_var *names
    names.each do |name|
      define_method name do |*val|
        return instance_variable_get("@#{name}") if val.empty?
        instance_variable_set "@#{name}", (val.size == 1)? val[0]: val
      end  
    end  
  end

  def def_hook name, *aliases
    # Workaround da define_method vor ruby 1.9 keine Blocks als Parameter unterstützt
    eval_methods name, "
      def #{name}(&block)
        ( @#{name} ||= Array.new ) << block if block_given?
        return @#{name}
      end", aliases
  end

  def def_listener name, *aliases
    # Workaround da define_method vor ruby 1.9 keine Blocks als Parameter unterstützt
    eval_methods name, "
      def #{name}(regex=/.*/,&block) 
        ( @#{name} ||= Array.new ) << [regex,block] if block_given?
        return @#{name}
      end", aliases
  end

private
  def eval_methods name, eval_string, aliases
    self.module_eval eval_string
    aliases.each do |al|
      alias_method al, name
    end
  end
end #ConfigurationHelper

class Configuration
  include Singleton
  extend ConfigurationHelper

  def self.define &block
    r = instance
    r.instance_eval &block
    return r
  end     

  def include mod
    self.class.class_eval { include mod } 
  end

  def_conf_var :port, :host, :user, :stop_at_signals
  def_listener :message_receiving, :mr
  def_listener :before_sending,   :bs
  def_listener :after_sending,    :as

  def_hook :on_startup
  def_hook :on_exit    
end #Configuration

class Chat 
  def initialize config_file = "config.rb"
    @config_file = config_file
  end

  def start
    load(File.join(File.dirname(__FILE__), @config_file), true)
    @config = Configuration.instance
    @socket = TCPSocket.new @config.host, @config.port
    @config.stop_at_signals.each { |signal| Signal.trap(signal) { stop } } unless @config.stop_at_signals.nil?
    @socket.puts "/nick #{@config.user}"
    @config.on_startup.each { |hook| hook.call @socket} unless @config.on_startup.nil?
    fork_input
    run
  end

  def stop
    @config.on_exit.each { |hook| hook.call @socket } unless @config.on_exit.nil?
    @socket.puts "/quit"
  end
  
private
  def fork_input
    # mkfifo leider nicht in ruby implementiert
    `mkfifo in` unless File.exists? "in"
    Thread.new do
      begin
        # syscall open blockiert process 
        result = %x[cat in]
        next if result.nil? or result == ""
        result = process @config.before_sending, result          
        @socket.puts result
        process @config.after_sending, result
      end until result.nil?
    end
  end

  def run
    loop { break unless process @config.message_receiving, @socket.gets }
  end

# helper
  def process listener, msg    
    unless listener.nil? or msg.nil?
      msg.delete!("\000")
      listener.each do |regex, hook|
        matches = msg.scan regex
        msg = hook.call(msg, matches, @socket) || msg unless matches.empty?
      end      
    end
    return msg
  end

end #Chat

chat = Chat.new "config.rb"
chat.start

